/* Copyright (c) 2009, Delft University of Technology

    This file is part of the MEGS package.

    The MEGS pakcage is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The MEGS pakcage is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the MEGS package.  If not, see
    <http://www.gnu.org/licenses/>.
*/
package engine;

import java.text.DecimalFormat;
import java.util.Vector;



public class Node implements Comparable{
	
	//used to keep track of the systems time
	protected SystemStatus system;
	
	//the node number
	protected int number;
	
	//last time an update was made
	protected double lastActivity = 0;
	
	//list of all messages sent/received
	protected Vector<BroadcastEvent> broadcastEvents = new Vector<BroadcastEvent>();
	protected Vector<SendSingleMsgEvent> sendEvents = new Vector<SendSingleMsgEvent>();
	protected Vector<ReceiveSendMsgEvent> receiveSendEvents = new Vector<ReceiveSendMsgEvent>();
	protected Vector<ReceiveBroadcastMsgEvent> receiveBcstEvents = new Vector<ReceiveBroadcastMsgEvent>();
	protected Vector<InterceptMsgEvent> interceptEvents = new Vector<InterceptMsgEvent>();
	
	/**
	 * Create a new node with adress/number n
	 * @param n
	 * 
	 */
	public Node (int n, SystemStatus s) 
	{
		number = n;
		system = s;
	}
	
	
	/**
	 * Sets last activity time to the given time. Updates the activity counter
	 * @param time
	 */
	public void recordActivity(double time)
	{
		lastActivity = time;
	}
	
	public double getLastActivity()
	{
		return lastActivity;
	}

	
	//compares the node numbers, if this is greater, then positive
	//Note: this class has a natural ordering that is inconsistent with equals.
	public int compareTo(Object o) {
		Node n = (Node) o;
		return this.number - n.number;
	}
	
	public int getNumber() {
		return number;
	}
	
	/**
	 * Calls idDead with the current system time
	 * @return
	 */
	public boolean isDead()
	{
		return isDead(system.getTime());
	}
	
	/**
	 * Returns true if a node hasn't reacted for MintRouteSystem.DEATH_CHECK
	 * seconds.
	 * @param systemTime The current system time (used to chekc how long this
	 * node hasn't reacted)
	 * @return
	 */
	public boolean isDead(double systemTime)
	{
		return getLastActivity() + Settings.DEATH_CHECK < systemTime;
	}
	
	/**
	 * Takes re and looks for a matching send event in this node
	 * @param re The receive event 
	 * @return The matching send event, or null if none found
	 */
	public SendMsgEvent findSendEvent(ReceiveMsgEvent re)
	{

		//find the earliest match 
		for(int i = 0; i < sendEvents.size(); i++)
		{
			if(Math.abs(sendEvents.elementAt(i).getTime() - re.getTime()) <  Settings.MESSAGE_TIME)
			{
				//check if the receiver is correct and if this event hasn't been matched yet
				if(sendEvents.elementAt(i).getToNode() == re.getNode() &&  !sendEvents.elementAt(i).matchedReceive())
				{
					return sendEvents.elementAt(i);
				}
			}
		}
		return null;
	}
	
	/**
	 * Takes re and looks for a matching intercepted send event in this node
	 * @param re The receive event 
	 * @return The matching send event, or null if none found
	 */
	public SendMsgEvent findInterceptedSendEvent(ReceiveMsgEvent re)
	{

		//find the earliest match 
		for(int i = 0; i < sendEvents.size(); i++)
		{
			if(Math.abs(sendEvents.elementAt(i).getTime() - re.getTime()) <  Settings.MESSAGE_TIME)
			{
			    //receiver is unimportant
				return sendEvents.elementAt(i);
			}
		}

		return null;
	}
	
	/**
	 * Retursn a broadcast event going with the given argument
	 * @param re
	 * @return
	 */
	public SendMsgEvent findBroadcastEvent(ReceiveBroadcastMsgEvent re)
	{

		
		for(int j = broadcastEvents.size() -1; j >= 0; j--)
		{
			
			if(Math.abs(broadcastEvents.elementAt(j).getTime() - re.getTime()) <  Settings.MESSAGE_TIME)
			{
				return broadcastEvents.elementAt(j);
			}
			
		}

		return null;
	}
	
	

	public Vector<BroadcastEvent> getBroadcastEvents() {
		return broadcastEvents;
	}


	public void setBroadcastEvents(Vector<BroadcastEvent> broadcastEvents) {
		this.broadcastEvents = broadcastEvents;
	}



	public Vector<SendSingleMsgEvent> getSendEvents() {
		return sendEvents;
	}


	public void setSendEvents(Vector<SendSingleMsgEvent> sendEvents) {
		this.sendEvents = sendEvents;
	}
	
	public String toString()
	{
		DecimalFormat df = new DecimalFormat("#.###");
		return number + " : last activity=" + df.format(lastActivity) +" s";  
	}


	public Vector<ReceiveBroadcastMsgEvent> getReceiveBcstEvents() {
		return receiveBcstEvents;
	}


	public Vector<ReceiveSendMsgEvent> getReceiveSendEvents() {
		return receiveSendEvents;
	}


	public Vector<InterceptMsgEvent> getInterceptEvents() {
		return interceptEvents;
	}


	public void setInterceptEvents(Vector<InterceptMsgEvent> interceptEvents) {
		this.interceptEvents = interceptEvents;
	}


	public void setReceiveBcstEvents(
			Vector<ReceiveBroadcastMsgEvent> receiveBcstEvents) {
		this.receiveBcstEvents = receiveBcstEvents;
	}


	public void setReceiveSendEvents(Vector<ReceiveSendMsgEvent> receiveSendEvents) {
		this.receiveSendEvents = receiveSendEvents;
	}
	
}
