/* Copyright (c) 2009, Delft University of Technology

    This file is part of the MEGS package.

    The MEGS pakcage is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The MEGS pakcage is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the MEGS package.  If not, see
    <http://www.gnu.org/licenses/>.
*/
package engine;

import java.util.Vector;
import java.io.*;

/**
 * Keeps track of the data and writes to a file that can be directly used
 * in programs like gnuplot or excel
 * # is the comment delimiter (gnuplot default)
 */
public class Plot {
	
	private String[] columns;
	private Vector<double[]> data = new Vector<double[]>();
	
	private double resolution = 0; //this is to prevent too many entries
	private int resolutionColumn = -1; //the column to check the resolution on, -1 indicates don't check
	/**
	 * Creates a new plot with the column names given (only for file description)
	 * @param colums
	 */
	public Plot(String[] columns)
	{
		this.columns = columns;
	}
	
	/**
	 * Makes a plot that only adds a value if a certain difference (res) on column (resCol) has been made.
	 * This is to prevent too many entries in the data file.
	 * This means that a data entry isn't added if the change in the resCol isn't bigger than res 
	 * @param columns
	 * @param res The resolution
	 * @param resCol The max column change
	 */
	public Plot(String[] columns, double res, int resCol)
	{
		this.columns = columns;
		resolution = res;
		resolutionColumn = resCol;
	}
	
	/**
	 * Add one data point, the length of data must be equal to the length of columns
	 * @param dat the data points to plot
	 */
	public void addPoint(double[] dat)
	{
		assert dat.length == columns.length : "Wrong data length";
		boolean add = true;
		//check the resolution
		if(resolutionColumn != -1)
		{
			if(data.size() > 0 && dat[resolutionColumn] < data.lastElement()[resolutionColumn] + resolution )
			{
				add = false;
			}
		}
		if(add)
		{
			data.add(dat);
		}
	}
	
	/**
	 * Write the data to the given file. If this file exists it will be overwritten
	 * @param f 
	 */
	public void writeData(File f) throws IOException
	{
		System.out.println("Writing plot to: " + f);
		
		FileWriter fw = new FileWriter(f);
		fw.write("Plot data by engine.Plot\n");
		fw.write("Columns:\n");
		for(String c : columns)
		{
			fw.write(c + "\t");
		}
		fw.write("\n");
		for(double[] dat : data)
		{
			for(double d : dat)
			{
				//Check for invalid numbers
				if(!Double.isNaN(d))
				{
					fw.write(d + "\t");
				}
				else
				{
					fw.write("0.0\t");
				}
				
			}
			fw.write("\n");
		}
		fw.close();
	}
}
