/* Copyright (c) 2009, Delft University of Technology

    This file is part of the MEGS package.

    The MEGS pakcage is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The MEGS pakcage is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the MEGS package.  If not, see
    <http://www.gnu.org/licenses/>.
*/
package engine;

import java.text.DecimalFormat;
import java.util.*;

import mintroute.MintNode;


/**
 * This class implements a lot of helper functions to provide the mechanism for
 * keeping track of the global state. 
 *
 * Many of these abstract methods take Node as an argument. This is done so these
 * methods can be called from general controlflow methods. Implementations
 * of these methods can safely cast to their own node type. As only one 
 * node type per implementation should be used. Classes extending this class
 * should make sure this stay true. (A work around for this cast is to take the node
 * id of the given node, and seach in the nodelist directly).
 * It might be possible to solve this with generics (T extends Node).
 * Of course the return type of any implemented node can be the specific class
 * instead of the general Node class to solve casting issues within the class.
 */
public abstract class SystemStatus{
	
	/**
	 * All the nodes in the system. 
	 */
	protected Vector<Node> nodes = new Vector<Node>();
	
	/**
	 * The list of all asserts, assert number with matching name,
	 * the constructor of a class extending this should fill this list 
	 */
	protected TreeMap<Integer, String> assertlist = new TreeMap<Integer, String>();
	
	/**
	 * The list of the asserts where the status is kept 
	 */
	protected Asserts asserts;
	
	/**
	 * The list of all the globalPredicates (asserts that are checked in a % of time instead of % of calls)
	 */
	protected Vector<GlobalPredicate> globalPredicates = new Vector<GlobalPredicate>();
	
	//vector holding all future asserts
	//These asserts need to be true before a certain time
	protected Vector<TimedAssert> timedAsserts = new Vector<TimedAssert>();
	
//	Errors have to be printed at the end aswell
	protected Vector<String> errors = new Vector<String>();
	

	
	/**
	 * The system time 
	 */
	protected double time = 0;
	
	public SystemStatus()
	{
		
	}
	
	/**
	 * This should be called once afte the object is constructed, this is 
	 * to allow constructor of the element to set things first.
	 */
	public void init()
	{
//		create the list of asserts (this reads the assert list)
		asserts = new Asserts(this);
	}
	
	/**
	 * Should be called once after all events are processed.
	 */
	public void end()
	{
		matchReceives();
	}
	
	
	/**
	 * This is called by the engine whenever a new event is found. 
	 * @param e The event found
	 */
	public void processEvent(Event e)
	{
		//assert time <= e.getTime() : "Found event with lower time than current event! ";
		//Check if the ordering worked properly (this has been replaced with an error)
		
		//update time now done at the end
		//time = e.getTime();
		
		Node n = getNode(e.getNode());
		updateTime(e.getTime());
		
		//node doesn't exist
		if(n == null)
		{
			n = addNode(e.getNode());
		}
		if(e instanceof AssertEvent)
		{
			doAssert(n, (AssertEvent) e);
		} 
		else if(e instanceof StatusEvent)
		{
			updateStatus(n, (StatusEvent) e);
			
		}
		else if(e instanceof BroadcastEvent)
		{
		
			n.getBroadcastEvents().add((BroadcastEvent)e);
		}
		else if(e instanceof ReceiveSendMsgEvent)
		{
			n.getReceiveSendEvents().add((ReceiveSendMsgEvent)e);

		}
		else if(e instanceof ReceiveBroadcastMsgEvent)
		{
			n.getReceiveBcstEvents().add((ReceiveBroadcastMsgEvent)e);

		}
		else if(e instanceof SendSingleMsgEvent)
		{
			n.getSendEvents().add((SendSingleMsgEvent)e);
		}
		else if(e instanceof InterceptMsgEvent)
		{
			n.getInterceptEvents().add((InterceptMsgEvent)e);

		}
		else if(e instanceof UnknownEvent)
		{
			//only report the time
			n.recordActivity(e.getTime());
		}	
		
	}
	
	/** Picks the right assert and evaluates the expression
	 * This method should be overwritten to allow 1 event to trigger multiple asserts
	 * @param n The node to do the assert on 
	 * @param e The assert event
	 */
	protected void doAssert(Node n, AssertEvent e)
	{
		executeAssert(n, e.getAssertNum());
	}
	
	/**
	 * Takes care of calling checkAssert and the administraction of its results
	 * This method shouldn't need to be overwritted
	 * @param n The node
	 * @param assertnum The number of the assert to be called
	 */
	protected void executeAssert(Node n, int assertnum)
	{
		//get the assert
		Assert as = asserts.getAssert(assertnum);
		try
		{
			boolean result = checkAssert(n, assertnum);
			if(result)
			{
				as.countSucces();
			}
			else
			{
				DecimalFormat df = new DecimalFormat("0.000");
				System.out.println("Assert FAILED: " + n.getNumber()  + " - " + df.format(time) + " : " + getAssertName(assertnum));
				as.countFailed();
			}
		}
		catch(AssertSkippedException e)
		{
			as.countSkipped();
		}
	}
		
	/**
	 * Called when an event updates status
	 * @param n The node (can safely be casted)
	 * @param e The event
	 */
	protected abstract void updateStatus(Node n, StatusEvent e);
	
	/**
	 * This should be called when changing the time, if the 
	 * time has changed this will call appropriate functions 
	 * @param newTime The new time
	 */
	protected void updateTime(double newTime)
	{
		//System.out.println ("Errors: " + errors.size() + " Timed asserts " + timedAsserts.size());
		
		//temp exit if many errors
		if(errors.size() > 1000)
		{
			System.out.println(errorReport() + "\n\nMore than 1000 errors so exit");
			
			System.exit(1);
		}
		
		//make sure time is in the correct order
		if(time > newTime)
		{
			String err = "Event at " + newTime + " with system time being " + time;
			errors.add(err);
		}
		
		if(newTime != time)
		{
//			make checks on the old time like globalPredicates
			newTime();
			checkTimedAsserts();
			
			//update the time
			time = newTime;
		}
	}
	
	/**
	 * Return the node with parameter n, or null if the node doesnt' exist
	 * @param n
	 * @return
	 */
	public Node getNode(int i)
	{
		for(Node n : nodes)
		{
			if(n.getNumber()==i)
			{
				return n;
			}
		}
		
		return null;
	}
	/**
	 * Make a new node of the type extending Node, with ID n
	 * @param n the node ID.
	 * @return The created node
	 */
	public abstract Node makeNode(int n);
	
	/**
	 * Add a node with node number n to the list of nodes, n should be unused
	 * @param n
	 * @return The created node
	 */
	public Node addNode(int i)
	{
		Node n = makeNode(i);
		nodes.add(n);
		
		System.out.println("Found new node: " + n);
		return n;
	}
	
	/**
	 * @return A vector with all nodes
	 */
	public Vector<Node> getAllNodes()
	{
		return nodes;
	}
	
	
	/**
	 * This method is called whenever the time has changed, this 
	 * method can be used to check globalPredicates for instance
	 * This method is called just before the time is updated
	 * By default this method does nothing.
	 */
	protected void newTime()
	{
		// only for overwriting
	}
	
	/**
	 * 
	 * Goes over all the active timed asserts and handles them
	 */
	public void checkTimedAsserts()
	{
		//System.out.println("Checking " + timedAsserts.size() + " timed asserts ");
		//remove these asserts after the iteration is done
		Vector<TimedAssert> remove = new Vector<TimedAssert>();
		
		for(TimedAssert ass : timedAsserts)
		{
			if(ass.isValid(time))
			{
				//check the assert
				try
				{
					boolean result = checkAssert(getNode(ass.getNode()), ass.getAssertNumber());
					if(result)
					{
						//remove the assert as it holds within time
						ass.assertResult(true);
						remove.add(ass);
					}
				}catch(AssertSkippedException e)
				{
					//No need to do anything, as a skipped assert is ignored here and if it
					//isn't fixed in time will count as false
				}
				
			}
			else
			{
				//too much time passed
				ass.assertResult(false);
				remove.add(ass);
			}
		}
		
		for(TimedAssert remass : remove)
		{
			timedAsserts.remove(remass);
		}
	}
	
	/**
	 * Checks the assert indicated by the parameters.
	 * A method implementing this method should cast n to 
	 * the subtype of n used by the system
	 * @param n The node to nbe checked
	 * @param num The number of the assert to check
	 * @return True if the assert holds, otherwise false
	 * @throws AssertSkippedException when the data isn't complete enuogh to execute the assert
	 */
	public abstract boolean checkAssert(Node n, int num) throws AssertSkippedException;
	
	
	
	public String getAssertStatus() {
		
		String s = asserts.toString() + "\nInvariants:\n";
		
		for(GlobalPredicate i : globalPredicates)
		{
			s+= i+"\n";
		}
		return s;
	}

	public abstract String fullReport();
	
	/**
	 * The number of asserts known in this system 
	 * @return
	 */
	public int numberOfAsserts() {
		
		return assertlist.size();
	}
	
	/**
	 * Returns the set with all the keys (Integer) in the asserts
	 * @return
	 */
	public Set<Integer> assertNums()
	{
		return assertlist.keySet();
	}
	
	/**
	 * Return the string representation of the assert going with num
	 * @param num
	 * @return
	 */
	public String getAssertName(int num)
	{
		return assertlist.get(num);
	}
	
	public double getTime()
	{
		return time;
	}
	
	/**
	 * Goes over the nodes and checks for activity within the last 'deadcheck' seconds
	 * @return returns the amounts of nodes that we haven't heard from in a while
	 */
	public int countDeadNodes()
	{
		int dead = 0;
		for(Node n : getAllNodes())
		{
			//check if node has responded within the bound to count as living
			if(n.isDead(time))
			{
				dead++;
			}
		}
		return dead;
	}
	
	/**
	 * Tries to match all receives with send events
	 */
	public void matchReceives()
	{
		
		
		//go over all receive events
		for(Node receiver : getAllNodes())
		{
			//The intercepts are done first, because the sending marks them as checked
//			go over all intercepted send events
			for(InterceptMsgEvent re : receiver.getInterceptEvents())
			{
				Node sender = getNode(re.getFromNode());
				//sender might not be known yet
				if(sender != null)
				{
					//don't ignore the messages marked 
					SendMsgEvent se = sender.findInterceptedSendEvent(re);
					if(se == null)
					{
						String err = "No send event found for: " + re;
						//check if node was dead at the time of the event
						if(sender.isDead(re.getTime()))
						{
							err += "(dead)";
							//optinally add this as a showed error by moving the add out of the else jsut below
							
						}
						else
						{
							if(Settings.SHOW_MESSAGE_ERRORS)
							{
								errors.add(err);
							}
						}
					}
					
				}
			}
			
			//go over all normal sent events
			for(ReceiveSendMsgEvent re : receiver.getReceiveSendEvents())
			{
				Node sender = getNode(re.getFromNode());
				//sender might not be known yet
				if(sender != null)
				{

					SendMsgEvent se = sender.findSendEvent(re);
					if(se == null)
					{
						String err = "No send event found for: " + re;
						//check if node was dead at the time of the event
						if(sender.isDead(re.getTime()))
						{
							err += "(dead)";
							//optinally add this as a showed error by moving the add out of the else jsut below
							
						}
						else
						{
							if(Settings.SHOW_MESSAGE_ERRORS)
							{
								errors.add(err);
							}
						}
					}
					else
					{
						se.eventMatched();
						//errors.add("Found send event: " +se + " for " + re);
					}
				}
				else
				{
			//		errors.add("No sender node found for: " + re);
				}
			}
			

//			go over all broadcast events
			for(ReceiveBroadcastMsgEvent re : receiver.getReceiveBcstEvents())
			{
				Node sender = getNode(re.getFromNode());
				//sender might not be known yet
				if(sender != null)
				{
					SendMsgEvent se = sender.findBroadcastEvent(re);
					if(se == null)
					{
						String err = "No send event found for: " + re;
						//check if node was dead at the time of the event
						if(sender.isDead(re.getTime()))
						{
							err += "(dead)";
							//optinally add this as a showed error by moving the add out of the else jsut below
							
						}
						else
						{
							if(Settings.SHOW_MESSAGE_ERRORS)
							{
								errors.add(err);
							}
						}
					}
					
				}
			}
		}
	}
	
	
	/**
	 * Returns a string representation of the errors in interpreting this log
	 * @return
	 */
	public String errorReport()
	{
		
		StringBuilder sb = new StringBuilder(40*errors.size());
		sb.append("Errors: \n");
		for(String e : errors)
		{
			sb.append(e);
			sb.append("\n");
		}
		sb.append(errors.size() + " errors");
	
		return sb.toString();
	}
}
